%{
	#include <stdio.h>
	#include "tb.h"
	extern int yylex();
	extern int lineNumber;
	
	extern vector<Producao*> listaProducoes;
	extern vector<Regra*> listaRegras;
	extern vector<string> listaTokens;
	
	int yyerror(char *msg)
	{
		fprintf(stderr, "Line %d: %s\n", lineNumber+1, msg);	
		exit(1);
		return 0;
		
	}
	
	extern int existeProducao(string nome);
	
	extern bool existeToken(string nome);

%}

%token IDENTIFIER
%token PC_LBRACE
%token PC_RBRACE
%token DOUBLE_PC
%token COLON
%token PC
%token PIPE
%token LBRACE
%token RBRACE
%token SEMICOLON
%token CHAR_DEF
%token PC_TOKEN

%type <opcoes> Opcoes
%type <atributos> Atributos
%type <str> IDENTIFIER
%type <chr> CHAR_DEF

%union {
	Opcoes *opcoes;
	Atributos *atributos;
	char *str;
	char chr;
}

%%

Tb : 
	Tokens DOUBLE_PC Producoes
	;

Tokens :
	Token
	| Tokens Token
	;

Token :
	PC_TOKEN IDENTIFIER 
	{
		listaTokens.push_back($2);
	}
	;

Producoes : 
	Producao 
	| Producao Producoes
	;
	
Producao : 
	IDENTIFIER COLON Opcoes SEMICOLON 
	{
		Regra *regra = new Regra;
		regra->nome = $1;
		regra->opcoes = $3;
		listaRegras.push_back(regra);
		
		int indexProducao = existeProducao($1);
		Producao *p = NULL;
		
		if(indexProducao == -1) {
			p = new Producao($1);
			listaProducoes.push_back(p);
		} else {
			p = listaProducoes[indexProducao];
		}
		
		//se tiver mais de uma opcao
		if ($3->opcoes.size() > 1) {
			for (int i=0; i<$3->opcoes.size(); i++) {
				//constroi o nome da opcao
				Atributos *op = $3->opcoes[i];
				
				string nome = "";
				for(int j=0; j < op->atributos.size();j++) {
					nome += op->atributos[j];
				}
				Producao *opcao = NULL;
				
				indexProducao = existeProducao(nome);
		
				if(indexProducao == -1) {
					if(nome == "") {
						nome = "Empty_";
						nome += $1;
					}
					opcao = new Producao(nome);
					if(op->atributos.size() > 1) {
						opcao->atributos = op->atributos;
					}
					listaProducoes.push_back(opcao);
				} else {
					opcao = listaProducoes[indexProducao];
				}
				opcao->pais.push_back($1);
			}
		} else {
			//se tiver uma opcao soh
			p->atributos = $3->opcoes[0]->atributos;
		}
	}
	;
	
Opcoes : 
	Atributos 
	{
		$$ = new Opcoes;
		$$->opcoes.push_back($1);
	}
	| Opcoes PIPE Atributos
	{
		$$ = new Opcoes;
		$$->opcoes = $1->opcoes;
		$$->opcoes.push_back($3);
	}
	| Opcoes PIPE
	{
		$$ = $1;
		$$->opcoes.push_back(new Atributos);
	}
	| PIPE Opcoes
	{
		$$ = new Opcoes;
		$$->opcoes.push_back(new Atributos);
		
		for(int i=0;i<$2->opcoes.size();i++) {
			$$->opcoes.push_back($2->opcoes[i]);
		}
	}
	;

Atributos : 
	IDENTIFIER 
	{ 
		$$ = new Atributos; 
		$$->atributos.push_back($1); 
	}
	| Atributos IDENTIFIER
	{
		$$ = new Atributos; 
		$$->atributos = $1->atributos; 
		$$->atributos.push_back($2);
	}
	;