#include "tb.h"
#include <fstream>
#include <stdlib.h>
#include <stdio.h>

using namespace std;

vector<Regra*> listaRegras;
vector<Producao*> listaProducoes;
vector<string> listaTokens;
extern int yyparse();

void imprimirProducao(Producao *p);
void gerarNos();
void gerarSaida(string arquivoEntrada);
void gerarRegra(Regra *regra, ofstream &out, bool isRoot);
void gerarAtributos(Atributos *atributos, ofstream &out, bool filhoUnico, string nomeRegra, bool isRoot);
void gerarVisitor();
bool existeToken(string nome);
int existeProducao(string nome);

int main(int argc, char **argv) {
	extern FILE *yyin;
	if(argc != 2) {
		printf("Uso: %s arquivo_de_entrada\n", argv[0]);
		exit(1);
	}
	
	if ((yyin = fopen(argv[1], "r")) == NULL) {
		fprintf(stderr, "Erro ao tentar abrir o arquivo %s!\n", argv[1]);
		exit(1);
	}
	
	if (yyparse()) {
		fprintf(stderr, "No foi possvel compilar %s!\n", argv[1]);
		exit(1);
	}
	printf("sucesso!\n\n");
	/*
	for(int i=0;i<listaTokens.size(); i++) {
		printf("%%token %s\n", listaTokens[i].c_str());
	}
	
	printf("\n\n");
	
	for(int i=0;i<listaProducoes.size(); i++) {
		imprimirProducao(listaProducoes[i]);
	}
	*/
	gerarNos();
	gerarSaida(argv[1]);
	gerarVisitor();
	return 0;
}



void imprimirProducao(Producao *p) {
	printf("Nome: %s\n", p->nome.c_str());
	printf("Pais: ");
	
	for(int i=0; i< p->pais.size(); i++) {
		printf("%s ", p->pais[i].c_str());
	}
	printf("\nAtributos: ");
	for(int i=0; i< p->atributos.size(); i++) {
		printf("%s ", p->atributos[i].c_str());
	}
	printf("\n\n");
}


bool existeToken(string nome) {
		
		for(int i=0; i<listaTokens.size();i++) {
			if(listaTokens[i] == nome) {
				return true;
			}
		}
		
		return false;
}

int existeProducao(string nome) {
	for(int i=0; i<listaProducoes.size();i++) {
		if(listaProducoes[i]->nome == nome) {
			return i;
		}
	}
	return -1;
}

void gerarNos() {
	/* Adiciona os Tokens na lista de producoes */
	for(int i=0; i<listaTokens.size(); i++) {
		if(existeProducao(listaTokens[i]) == -1) {
			Producao *p = new Producao(listaTokens[i]);
			listaProducoes.push_back(p);
		}
	}
	
	//gera os ns
	for(int i=0; i<listaProducoes.size(); i++) {
		Producao *p = listaProducoes[i];
		bool isToken = false;
		string nomeArquivo = "";
		string ifndef = "";
		
		if(existeToken(p->nome)) {
			p->atributos.push_back("#valor");
			p->atributos.push_back("#codigo");
			nomeArquivo += "Token_";
			ifndef = "Token_";
			isToken = true;
		}
		nomeArquivo += p->nome + ".h";
		ifndef += p->nome;
		ofstream out(nomeArquivo.c_str());
		
		out << "/* Arquivo gerado pelo tb */\n\n";
		
		for(int j=0;j<ifndef.size();j++) {
			ifndef[j] = toupper(ifndef[j]);
		}
		out << "#ifndef " << ifndef << "_H\n";
		out << "#define " << ifndef << "_H\n\n";
		 
		if(p->pais.size() > 0) {
			for(int j=0;j<p->pais.size(); j++) {
				out << "#include \"" << p->pais[j] << ".h\"\n";
			}
		}
		
		if(isToken) {
			out << "#include <string>\n";
			string str = "Token_";
			p->nome = str + p->nome;
		}
		
		out << "\n\n";
		
		for(int j=0; j<p->atributos.size();j++) {
			string str = "";
			if (existeToken(p->atributos[j])){
				str = "Token_";
			}
			if(p->atributos[j][0] != '#')
				out << "class " << str + p->atributos[j] << ";\n";
		}
		out << "class Visitor;\n";
		
		out << "\nusing namespace std;\n\n\nclass " << p->nome;
		
		if(p->pais.size() > 0) {
			out << " : public " << p->pais[0];
			
			for(int j=1;j<p->pais.size(); j++) {
				out << ", public " << p->pais[j];
			}
		}
		out << " {\n";
		out << "\tprivate:\n";
		for(int j=0; j<p->atributos.size();j++) {
			if(p->atributos[j] == "#valor") {
				out << "\t\tstring valor;\n";
			} else if(p->atributos[j] == "#codigo") {
				out << "\t\tint codigo;\n";
			} else {
				string str = "";
				if (existeToken(p->atributos[j])){
					str = "Token_";
				}
				out << "\t\t" << str + p->atributos[j] << " *" << p->atributos[j] << "_;\n";
			}
		}
		out << "\tpublic:\n";
		out << "\t\t" << p->nome << "(";
		for(int j=0; j<p->atributos.size();j++) {
			if(p->atributos[j] == "#valor") {
				out << "string valor";
			} else if(p->atributos[j] == "#codigo") {
				out << "int codigo";
			} else {
				string str = "";
				if (existeToken(p->atributos[j])){
					str = "Token_";
				}
				out << str + p->atributos[j] << " *" << p->atributos[j] << "_";
			}
			if(j+1 != p->atributos.size()) {
				out << ", ";
			}
		}
		out << ")";
		if(p->atributos.size() > 0) {
			out << " : ";
			for(int j=0; j<p->atributos.size();j++) {
				if(p->atributos[j] == "#valor") {
					out << "valor(valor)";
				} else if(p->atributos[j] == "#codigo") {
					out << "codigo(codigo)";
				} else {
					out << p->atributos[j] << "_(" << p->atributos[j] << "_)";
				}
				if(j+1 != p->atributos.size()) {
					out << ", ";
				}
			}
		}
		out << " {}\n\n";
		
		for(int j=0; j<p->atributos.size();j++) {
			if(p->atributos[j] == "#valor") {
				out << "\t\tstring getValor() {\n\t\t\treturn valor;\n\t\t}\n\n";
			} else if(p->atributos[j] == "#codigo") {
				out << "\t\tint getCodigo() {\n\t\t\treturn codigo;\n\t\t}\n\n";
			} else {
				string str = "";
				if (existeToken(p->atributos[j])){
					str = "Token_";
				}
				out << "\t\t" << str + p->atributos[j] << " *get" << p->atributos[j] << "() {\n";
				out << "\t\t\t return " << p->atributos[j] << "_;\n";
				out << "\t\t}\n\n";
			}
		}
		out << "\t\tvirtual void *visit(Visitor *visitor);\n";
		out << "};\n\n";
		out << "#endif\n";
		
		string nomeArquivoCpp = nomeArquivo;
		nomeArquivoCpp.erase(nomeArquivoCpp.size() - 1);
		nomeArquivoCpp += "cpp";
		out.close();
		out.open(nomeArquivoCpp.c_str());
		out << "#include \"" << p->nome << ".h\"\n\n";
		out << "#include \"Visitor.h\"\n";
		out << "void *" << p->nome << "::visit(Visitor *visitor) {\n";
		out << "\treturn visitor->visit" << p->nome << "(this);\n";
		out << "}\n";
		
		out.close();
	}
}

void gerarSaida(string arquivoEntrada) {
	ofstream out((arquivoEntrada + ".tb").c_str());
	
	out << "%{\n";
	for(int i=0;i<listaProducoes.size();i++) {
		out << "#include \"" << listaProducoes[i]->nome << ".h\"\n";
	}
	out << "\nexternal " << listaRegras[0]->nome << " *root;\n";
	out << "%}\n\n";
	
	for(int i=0;i<listaTokens.size(); i++) {
		out << "%token " << listaTokens[i] << "\n";
	}
	
	out << "\n";
	
	for(int i=0;i<listaRegras.size();i++) {
		Regra *regra = listaRegras[i];
		string tipoRegra = regra->nome;
		tipoRegra[0] = tolower(tipoRegra[0]);
		out << "%type <" << tipoRegra << "> " << regra->nome << "\n";
	}
	out << "\n%union {\n";
	for(int i=0;i<listaRegras.size();i++) {
		Regra *regra = listaRegras[i];
		string tipoRegra = regra->nome;
		tipoRegra[0] = tolower(tipoRegra[0]);
		out << "\t" << regra->nome << " *" << tipoRegra << ";\n";
	}
	out << "}\n\n%%\n\n";
	
	for(int i=0;i<listaRegras.size();i++) {
		gerarRegra(listaRegras[i], out, i == 0);
	}
}

void gerarRegra(Regra *regra, ofstream &out, bool isRoot) {
	out << regra->nome << " : \n\t";
	Opcoes *op = regra->opcoes;
	for(int i=0;i<op->opcoes.size();i++) {
		
		gerarAtributos(op->opcoes[i], out, op->opcoes.size() == 1, regra->nome, isRoot);
		if(i+1 != op->opcoes.size()) {
			out << "\t| ";
		}
	}
	out << "\n;\n\n";
}


void gerarAtributos(Atributos *atributos, ofstream &out, bool filhoUnico, string nomeRegra, bool isRoot) {
	for(int i=0;i<atributos->atributos.size();i++) {
		out << atributos->atributos[i];
		if(i+1 != atributos->atributos.size()) {
			out << " ";
		}
	}
	out << "\n\t{\n\t\t";
	
	string root;
	if(isRoot) {
		root = "root";
	} else {
		root = "$$";
	}
	
	if(atributos->atributos.size() == 1) {
		if(existeToken(atributos->atributos[0])) {
			out << root << " = new Token_" << atributos->atributos[0] << "($1, " << atributos->atributos[0] << ");\n";
		} else {
			out << root << " = $1;\n";
		}
	} else {
		string nomeTipo = "";
		string param = "(";
		for(int i=0;i<atributos->atributos.size();i++) {
			nomeTipo += atributos->atributos[i];
			char num[80];
			if(existeToken(atributos->atributos[i])) {
				sprintf(num, "new Token_%s($%d, %s)", atributos->atributos[i].c_str(), i+1, atributos->atributos[i].c_str());
			} else {
				sprintf(num, "$%d", i+1);
			}
			param += num;
			if(i+1 != atributos->atributos.size()) {
				param += ", ";
			}
		}
		param += ")";
		if(nomeTipo == "") {
			nomeTipo = "Empty_";
			nomeTipo += nomeRegra;
		}
		if(filhoUnico) {
			nomeTipo = nomeRegra;
		}
		out << root << " = new " << nomeTipo << param << ";\n";
	}
	out << "\t}\n";
}


void gerarVisitor() {
	ofstream out("Visitor.h");
	out << "#ifndef VISITOR_H\n";
	out << "#define VISITOR_H\n\n";
	for(int i=0;i<listaProducoes.size();i++) {
		out << "#include \"" << listaProducoes[i]->nome << ".h\"\n";
	}
	out << "\n\nclass Visitor {\n";
	out << "\tpublic:\n";
	for(int i=0;i<listaProducoes.size();i++) {
		Producao *p = listaProducoes[i];
		out << "\t\t/**\n";
		out << "\t\t *\t" << p->nome << ":\n\t\t *\n";
		
		for(int j=0; j<p->atributos.size();j++) {
			if(p->atributos[j] == "#valor") {
				out << "\t\t *\tstring getValor();\n";
			} else if(p->atributos[j] == "#codigo") {
				out << "\t\t *\tint getCodigo();\n";
			} else {
				string str = "";
				if (existeToken(p->atributos[j])){
					str = "Token_";
				}
				out << "\t\t *\t" << str + p->atributos[j] << " *get" << p->atributos[j] << "();\n";
			}
		}
		out << "\t\t */\n";
		string param = p->nome;
		param[0] = tolower(param[0]);
		out << "\t\tvirtual void *visit"<< p->nome << "(" 
			 << p->nome << " *" << param << ") {}\n";
	}
	out << "};\n";
	out << "#endif\n";
}
